library A1216e;

uses
  Windows, SysUtils, Classes, WinSvc, ACCES32;

{$R *.RES}

type TA1216E_InitData = record
    BaseAddress: DWORD;
    IRQ: DWORD;
    NumPoints: LongInt;
    BusType: LongInt;
    BusNumber: DWORD;
end;

type PSingle = array [0..1] of Single;

var
    hDevice : HWND;
    EnvBuf : array[0..31] of char;
    ManagerHandle, ServiceHandle : SC_HANDLE;
    OS : Byte;

    BaseAddress: DWORD;
    NumPoints: LongInt;
    Rate: Double;
    DivisorA, DivisorB: WORD;
    G: array [0..255] of Byte;
//    RG: array [0..7] of Word;

const
    ServiceExe: PChar = '%SystemRoot%\SYSTEM32\DRIVERS\A1216E.sys';
    // control codes created by CTL_CODE macro in winioctl.h
    InitDriver = $80FF2004;
    Status = $80FF2008;
    Stop = $80FF200C;
    GetData = $80FF2010;
    InitChannelTable = $80FF2014;
    InitGainTable = $80FF2018;

    DriverName : PChar = 'A1216E';
    Args : PChar = NIL;
    kWin = 1;
    kNT = 2;
    //   the following are defined in winnt.h
    SERVICE_AUTO_START = $02;
    SERVICE_DEMAND_START = $03;
    SERVICE_ERROR_NORMAL = $01;
    SERVICE_KERNEL_DRIVER = $01;

    RG: array [0..3] of Integer = (1, 10, 100, 1000);

procedure LoadCounters;
begin
    { set counters 1 and 2 to mode 2 }
	OutPortB(BaseAddress + $F, $74);
	OutPortB(BaseAddress + $F, $B4);
    { load counters w/ divisors }
	OutPortB(BaseAddress + $D, DivisorA and $FF);
    OutPortB(BaseAddress + $D, (DivisorA shr 8) and $FF);
	OutPortB(BaseAddress + $E, DivisorB and $FF);
    OutPortB(BaseAddress + $E, (DivisorB shr 8) and $FF);
end;

function A1216E_Stop: LongInt; cdecl;
var
    cbRet: Integer;
begin
    if DeviceIoControl(hDevice, Stop, nil, 0, nil, 0, cbRet, nil) then
        Result := 0
    else
        Result := -1;
end;

function A1216E_Setup(theBase: DWORD; theIRQ: BYTE; theBusType: SmallInt; theBusNumber: WORD;
            theChannelTable: Pointer; theCTSize: DWORD; theGainTable: Pointer; theGTSize: DWORD;
            var theRate: Double; theNumPoints: LongInt): LongInt; cdecl;
var
    InitData: TA1216E_InitData;
    cbRet, i, best: Integer;
    Success: Boolean;
    Counts, d, temp, lowest, dDivA, dDivB: Double;
    SqDivisor: Word;
begin
    BaseAddress := theBase;

    if theRate > 0 then
        Counts := 1000000 / theRate
    else
        Counts := 65535.0 * 65535.0;
    if Counts > 65535.0 * 65535.0 then
        Counts := 65535.0 * 65535.0;
    if Counts < 4 then
        Counts := 4;
    SqDivisor := Trunc(sqrt(Counts));
    best := 2;
    lowest := 65535.0;
    for i := 2 to SqDivisor do
    begin
        d := i;
        temp := Abs(Counts - (d * Round(Counts / d)));
        if (temp < lowest) and (Round(Counts / i) <= 65535) then
        begin
            lowest := temp;
            best := i;
            if lowest = 0 then
                Break;
        end;
    end;
    DivisorB := best;
    DivisorA := Round(Counts / DivisorB);
    dDivA := DivisorA;
    dDivB := DivisorB;
    Rate := 1000000.0 / (dDivA * dDivB);
    theRate := Rate;

    NumPoints := theNumPoints;

    InitData.BaseAddress := BaseAddress;
    InitData.IRQ := theIRQ;
    InitData.NumPoints := NumPoints;
    InitData.BusType := theBusType;
    InitData.BusNumber := theBusNumber;

    MoveMemory(@G, theGainTable, theGTSize);

    Success := DeviceIoControl(hDevice, InitGainTable, theGainTable, theGTSize, nil, 0, cbRet, nil);
    if Success then
        Success := DeviceIoControl(hDevice, InitChannelTable, theChannelTable, theCTSize, nil, 0, cbRet, nil);
    if Success then
        Success := DeviceIoControl(hDevice, InitDriver, @InitData, sizeof(InitData), nil, 0, cbRet, nil);
    if Success then
        Result := 0
    else
        Result := -1;
end;

function A1216E_Go: LongInt; cdecl;
begin
    if BaseAddress <> 0 then
    begin
        LoadCounters;
        Result := 0;
    end else
        Result := -1;
end;

function A1216E_Status: LongInt; cdecl;
var
    Remaining: LongInt;
    cbRet: Integer;
begin
    if DeviceIoControl(hDevice, Status, nil, 0, @Remaining, sizeof(Remaining), cbRet, nil) then
        Result := Remaining
    else
        Result := -1;
end;

function A1216E_GetData(Buf: Pointer; BufSize: DWORD; Unipolar: Byte; Range: Byte): LongInt; cdecl;
var
    i: Integer;
    cbRet: LongInt;
    sv, sch: Single;
    Counts: LongInt;
    Channel: DWORD;
    x: DWORD;
const
    Boundary: Integer = 8;
begin
    if DeviceIoControl(hDevice, GetData, nil, 0, Buf, BufSize, cbRet, nil) then
    begin
        for i := 0 to (cbRet div Boundary) - 1 do
        begin
            MoveMemory(@x, Pointer(DWORD(Buf) + (i * Boundary)), 4);
            Counts := x and $FFFF;
            Channel := (x shr 16) and $FFFF;
            if Unipolar > 0 then
                sv := (Range * Counts) / (4095 * RG[G[Channel]])
            else
                sv := (Range * (Counts - 2048)) / (2047 * RG[G[Channel]]);
            sch := Channel;
            MoveMemory(Pointer(DWORD(Buf) + (i * Boundary)), @sch, Boundary div 2);
            MoveMemory(Pointer(DWORD(Buf) + (i * Boundary) + (Boundary div 2)), @sv, Boundary div 2);
        end;
        Result := cbRet;
    end else
        Result := -1;
end;

{ Visual Basic interface }

function VBA1216E_Setup(theBase: DWORD; theIRQ: BYTE; theBusType: SmallInt; theBusNumber: WORD;
            theChannelTable: Pointer; theCTSize: DWORD; theGainTable: Pointer; theGTSize: DWORD;
            var theRate: Double; theNumPoints: LongInt): LongInt; cdecl;
begin
    Result := A1216E_Setup(theBase, theIRQ, theBusType, theBusNumber, theChannelTable, theCTSize,
                          theGainTable, theGTSize, theRate, theNumPoints);
end;

function VBA1216E_Go: LongInt; stdcall;
begin
    Result := A1216E_Go;
end;

function VBA1216E_Status: LongInt; stdcall;
begin
    Result := A1216E_Status;
end;

function VBA1216E_Stop: LongInt; stdcall;
begin
    Result := A1216E_Stop;
end;

function VBA1216E_GetData(Buf: Pointer; cbSize: DWORD; Unipolar: Byte; Range: Byte): LongInt; cdecl;
begin
    Result := A1216E_GetData(Buf, cbSize, Unipolar, Range);
end;

exports
    A1216E_Setup,
    A1216E_Setup name '_A1216E_Setup',
    A1216E_Go,
    A1216E_Go name '_A1216E_Go',
    A1216E_Status,
    A1216E_Status name '_A1216E_Status',
    A1216E_Stop,
    A1216E_Stop name '_A1216E_Stop',
    A1216E_GetData,
    A1216E_GetData name '_A1216E_GetData',
    // the following make Visual Basic happy
    VBA1216E_Setup, VBA1216E_Go, VBA1216E_Status, VBA1216E_Stop, VBA1216E_GetData;

procedure DLLHandler(Reason: Integer);
var
   ServiceStatus : TServiceStatus;
begin
    if (Reason = DLL_PROCESS_DETACH) and (hDevice <> 0) then
    begin
        CloseHandle(hDevice);
        Exit; //if auto start
        if OS = kNT then
        begin
            ManagerHandle := OpenSCManager (NIL, NIL, SC_MANAGER_ALL_ACCESS);
            ServiceHandle := OpenService(ManagerHandle, DriverName, SERVICE_ALL_ACCESS);
            ControlService(ServiceHandle, SERVICE_CONTROL_STOP, ServiceStatus);
            CloseServiceHandle(ManagerHandle);
            CloseServiceHandle(ServiceHandle);
        end;
    end;
end;

begin
    DLLProc := @DLLHandler;
    if hDevice = 0 then
    begin
        OS := kWin;
        GetEnvironmentVariable('OS', EnvBuf, 32);
        if (StrIComp(EnvBuf, 'WINDOWS_NT') = 0) then
        begin
            OS := kNT;
            // open manager
            ManagerHandle := OpenSCManager (NIL, NIL, SC_MANAGER_ALL_ACCESS);
            // create service
            ServiceHandle := CreateService (ManagerHandle,          // SCManager database
                                           DriverName,           // name of service
                                           DriverName,           // name to display
                                           SERVICE_ALL_ACCESS,    // desired access
                                           SERVICE_KERNEL_DRIVER, // service type
                                           SERVICE_AUTO_START,  // start type
                                           SERVICE_ERROR_NORMAL,  // error control type
                                           ServiceExe,            // service's binary
                                           NIL,                  // no load ordering group
                                           NIL,                  // no tag identifier
                                           NIL,                  // no dependencies
                                           NIL,                  // LocalSystem account
                                           NIL                   // no password
                                           );

            // open if already created
            if (ServiceHandle = 0) then
                ServiceHandle := OpenService(ManagerHandle, DriverName, SERVICE_ALL_ACCESS);

            StartService (ServiceHandle, 0, Args);

            CloseServiceHandle (ServiceHandle);
            CloseServiceHandle (ManagerHandle);
            hDevice := CreateFile('\\.\A1216E',
                              GENERIC_READ or GENERIC_WRITE,
                              FILE_SHARE_READ or FILE_SHARE_WRITE,
                              nil,
                              OPEN_EXISTING,
                              0,
                              0)
        end else
            hDevice := CreateFile('\\.\A1216E.vxd',
                                  GENERIC_READ or GENERIC_WRITE,
                                  FILE_SHARE_READ or FILE_SHARE_WRITE,
                                  nil,
                                  OPEN_EXISTING,
                                  FILE_FLAG_DELETE_ON_CLOSE,
                                  0)
    end;
end.
